
/**************************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2019-2019. All rights reserved.
 * Description: Rte_LaneDepartureWarning.c - The file has been not generated and validated in accordance with a safety standard
 * Author: VOS OS
 * Create: 2019-12-31
 *
 **************************************************************************************/

/***************************************************************************
 *
 * Includes
 *
 **************************************************************************/

/*
 * To restrict the SW-C APIs to a runnable use ONE of the following macros
 * #define RTE_RUNNABLEAPI_LDW_Controller
 */

/* Include application header file */
#include "Rte_LaneDepartureWarning.h"
/* Version verification of the Rte_LaneDepartureWarning module */
#if ((RTE_LANEDEPARTUREWARNING_AR_RELEASE_MAJOR_VERSION != 4U) || (RTE_LANEDEPARTUREWARNING_AR_RELEASE_MINOR_VERSION != 4U))
#error The imported RTE_LANEDEPARTUREWARNING module version is incompatible, there may be problems.
#endif


// Start of user code for user-defined code

// End of user code
/********************************************************************************************************************
 * AtomicSwComponentType : LaneDepartureWarning
 * ------------------------------------------------------------------------------------------------------------------
 *
 *******************************************************************************************************************/
#define LANEDEPARTUREWARNING_START_SEC_CODE
#include "LaneDepartureWarning_MemMap.h"
/********************************************************************************************************************
 * Runnable Entity : RLaneDepartureWarning
 * ------------------------------------------------------------------------------------------------------------------
 *
 * =======================================
 *           Sender-Receiver
 * =======================================
 * @Function name : Rte_DRead_LanesPosition_OppositeLaneDistance
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_RoadSegment_CurvatureRadius
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_RoadSegment_CurveCenterX
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_RoadSegment_CurveCenterY
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_LanesPosition_RoadEndLeftDistance
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_LanesPosition_RoadEndRightDistance
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_RoadSegment_Length
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_SteeringPosition_SteeringPosition
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(sint16, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_VehicleSpeed_VehicleSpeed
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_Write_Warning_Severity
 * @Description   : Initiate an "explicit" sender-receiver transmission of data elements with "data" semantic
 *                  (swImplPolicy different from queued).
 * @Parameter     : VAR(uint8, AUTOMATIC) value
 * @Return        : VAR(Std_ReturnType, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 *
 *******************************************************************************************************************/

/********************************************************************************************************************
 * Start of Section for user's comment
 ********************************************************************************************************************/
// Start of user code : Add user comment for RLaneDepartureWarning

#include <math.h>

#define WHEELS_DISTANCE             2U

#define MINIMUM_OPERATING_SPEED     60.0

#define SEVERITY_NONE       0
#define SEVERITY_LOW        1
#define SEVERITY_MEDIUM     2
#define SEVERITY_HIGH       3
#define SEVERITY_CRITICAL   4

/**
 * @brief Checks whether a depature point (dx, dy) is inside the road segment.
 *
 * The departure point (dx, dy) belongs to the circumference ((x - cx)^2 + (y - cy)^2 = r^2).
 *
 * This method computes the starting point (sx, sy) of the road segment by resolving the system
 *
 * (x - cx)^2 + (y - cy)^2 = r^2
 * x = (cx/cy)y
 *
 * where x = (cx/cy)y is the straight line that crosses the road curve center (cx, cy) and the vehicle (0,0).
 * This method computes the distance between (sx, sy) and (dx, dy) and compares it against the road segment length.
 *
 * @param dx Departure point (X).
 * @param dy Departure point (Y).
 * @param cx Road curve center (X).
 * @param cy Road curve center (Y).
 * @param r Road curvature radius w.r.t. the departure lane.
 * @param rs Road curvature radius w.r.t. the vehicle.
 * @param l Distance of the vehicle from the front end of the road segment (m).
 * @return 1 if the departure point is inside the road segment, 0 otherwise.
 */
boolean inRoadSegment(float64 dx, float64 dy, float64 cx, float64 cy, float64 r, float64 rs, float64 l)
{
    /* compute the intersection points between (x - cx)^2 + (y - cy)^2 = r^2 and x = (cx/cy)y */
    float64 a;
    float64 b;
    float64 c;
    float64 delta;

    a = 1 + ((cx * cx) / (cy * cy));
    b = -((2 * ((cx * cx) / cy)) + (2 * cy));
    c = (cx * cx) + (cy * cy) - (r * r);
    delta = (b * b) - (4 * a * c);
    if (delta > 0) /* there should always be solutions */
    {
        float64 d;  /* distance between the starting point of the segment (sx, sy) and the departure point (dx, dy) */
        float64 x1; /* intersection point 1 (X) */
        float64 y1; /* intersection point 1 (Y) */
        float64 d1; /* distance between the intersection point (x1, y1) and the vehicle (0, 0) */
        float64 x2; /* intersection point 2 (X) */
        float64 y2; /* intersection point 2 (Y) */
        float64 d2; /* distance between the intersection point (x2, y2) and the vehicle (0, 0) */
        float64 sx; /* starting point of the road segment (X) */
        float64 sy; /* starting point of the road segment (Y) */
        float64 lr; /* length of the arc between the starting point of the segment (sx, sy) and the departure point (dx, dy) */

        y1 = ((-b) - sqrt(delta)) / (2 * a);
        x1 = (cx / cy) * y1;
        d1 = sqrt( (x1 * x1) + (y1 * y1) );

        y2 = ((-b) + sqrt(delta)) / (2 * a);
        x2 = (cx / cy) * y2;
        d2 = sqrt( (x2 * x2) + (y2 * y2) );

        /* consider the intersection point with the shortest distance from the vehicle */
        sx = d1 < d2 ? x1 : x2;
        sy = d1 < d2 ? y1 : y2;
        d = sqrt(((dx - sx) * (dx - sx)) + ((dy - sy) * (dy - sy)));

        lr = 2 * r * asin(d / (2 * r));

        /* (l / rs) is the angle the vehicle runs to reach the front end of the segment */
        /* ((l / rs) * r) is the length of the segment (w.r.t. the circumference where the departure point is located) */
        return lr <= ((l / rs) * r);
    }
    return 0;
}

/**
 * @brief Computes the time that the vehicle takes to reach the departure point.
 *
 * @param dx Departure point (X).
 * @param dy Departure point (Y).
 * @param rv Vehicle curvature radius (0 if the vehicle is going straight)
 * @param v Vehicle speed (km/h).
 * @return The time of departure (s).
 */
float64 timeOfDeparture(float64 dx, float64 dy, float64 rv, float64 v)
{
    float64 d; /* distance between the vehicle (0,0) and the departure point (dx,dy) */
    float64 t; /* time of departure */

    d = sqrt((dx * dx) + (dy * dy));
    if (rv > 0)
    {
        float64 b; /* angle the vehicle runs to reach the departure point (dx, dy) */
        float64 a; /* angular speed */

        b = 2 * asin(d / (2 * rv));
        a = (v * 1000/3600) / rv;
        t = b / a;
    }
    else /* vehicle is going straight */
    {
        t = d / (v * 1000/3600);
    }
    return t;
}

/**
 * @brief Computes the departure time that a curving vehicle takes to cross a straight lane (x = k).
 *
 * This method computes the departure points (k, y1) and (k, y2) by resolving the system
 *
 * (x - vx)^2 + (y - vy)^2 = rv^2
 * x = k
 *
 * @param vx Vehicle curve center (X).
 * @param vy Vehicle curve center (Y).
 * @param rv Vehicle curvature radius.
 * @param k Lane position (X) w.r.t. the vehicle (0,0).
 * @param l Distance of the vehicle from the front end of the road segment (m).
 * @param v Vehicle speed (km/h).
 * @return Time of departure (0 means no departure).
 */
float64 timeToCurvingCrossLane(float64 vx, float64 vy, float64 rv, float64 k, float64 l, float64 v)
{
    /* compute the intersection points between (x - vx)^2 + (y - vy)^2 = rv^2 and x = k */
    float64 a;
    float64 b;
    float64 c;
    float64 delta;
    float64 y;  /* intersection point (Y) */
    float64 td; /* time of departure */

    td = 0;

    a = 1;
    b = -(2 * vy);
    c = (k * k) - (2 * k * vx) + (vx * vx) + (vy * vy) - (rv * rv);
    delta = (b * b) - (4 * a * c);
    if (delta > 0)
    {
        y = ((-b) - sqrt(delta)) / (2 * a);
        if (y <= 0) /* discard solutions behind the vehicle */
        {
            y = ((-b) + sqrt(delta)) / (2 * a);
        }
        if (y > 0 && y <= l) /* consider only solutions inside the road segment */
        {
            td = timeOfDeparture(k, y, rv, v);
        }
    }
    return td;
}

/**
 * @brief Computes the departure time that a going straight vehicle (x = 0) takes to cross a curving lane.
 *
 * This method computes the departure points (0, y1) and (0, y2) by resolving the system
 *
 * (x - cx)^2 + (y - cy)^2 = r^2
 * x = 0
 *
 * @param cx Road curve center (X).
 * @param cy Road curve center (Y).
 * @param r Road curvature radius.
 * @param rs Road curvature radius w.r.t. the vehicle.
 * @param l Distance of the vehicle from the front end of the road segment (m).
 * @param v Vehicle speed (km/h).
 * @return Time of departure (0 means no departure).
 */
float64 timeToCrossCurvingLane(float64 cx, float64 cy, float64 r, float64 rs, float64 l, float64 v)
{
    /* compute the intersection points between (x - cx)^2 + (y - cy)^2 = r^2 and x =0 */
    float64 a;
    float64 b;
    float64 c;
    float64 delta;
    float64 y;  /* intersection point (Y) */
    float64 td; /* time of departure */

    td = 0;

    a = 1;
    b = -(2 * cy);
    c = (cx * cx) + (cy * cy) - (r * r);
    delta = (b * b) - (4 * a * c);
    if (delta > 0)
    {
        y = ((-b) - sqrt(delta)) / (2 * a);
        if (y <= 0) /* discard solutions behind the vehicle */
        {
            y = ((-b) + sqrt(delta)) / (2 * a);
        }
        if (y > 0 && inRoadSegment(0, y, cx, cy, r, rs, l)) /* consider only solutions inside the road segment */
        {
            td = timeOfDeparture(0, y, 0, v);
        }
    }
    return td;
}

/**
 * @brief Computes the departure time that a curving vehicle takes to cross a curving lane.
 *
 * This method computes the departure points (x1, y1) and (x2, y2) by resolving the system
 *
 * (x - cx)^2 + (y - cy)^2 = r^2
 * (x - vx)^2 + (y - vy)^2 = rv^2
 *
 * which is equivalent to the system among the first circumference and the radical axis (i.e., first circumference - second circumference)
 *
 * (x - cx)^2 + (y - cy)^2 = r^2
 * ((x - cx)^2 + (y - cy)^2 = r^2) - ((x - vx)^2 + (y - vy)^2 = rv^2) = 0
 *
 * @param cx Road curve center (X).
 * @param cy Road curve center (Y).
 * @param r Road curvature radius.
 * @param vx Vehicle curve center (X).
 * @param vy Vehicle curve center (Y).
 * @param rv Vehicle curvature radius.
 * @param rs Road curvature radius w.r.t. the vehicle.
 * @param l Distance of the vehicle from the front end of the road segment (m).
 * @param v Vehicle speed (km/h).
 * @return Time of departure (0 means no departure).
 */
float64 timeToCurvingCrossCurvingLane(float64 cx, float64 cy, float64 r, float64 vx, float64 vy, float64 rv, float64 rs, float64 l, float64 v)
{
    /* compute the coefficients of the radical axis (ax + by + c = 0) */
    float64 a;
    float64 b;
    float64 c;

    a = (2 * vx) - (2 * cx);
    b = (2 * vy) - (2 * cy);
    c = (cx * cx) + (cy * cy) - (r * r) - (vx * vx) - (vy * vy) + (rv * rv);

    float64 aa;
    float64 bb;
    float64 cc;
    float64 delta;
    float64 x1;  /* intersection point 1 (X) */
    float64 y1;  /* intersection point 1 (Y) */
    float64 td1; /* time of departure of (x1, y1) */
    float64 x2;  /* intersection point 2 (X) */
    float64 y2;  /* intersection point 2 (Y) */
    float64 td2; /* time of departure of (x2, y2) */

    td1 = 0;
    td2 = 0;

    /* compute the intersection points between (x - cx)^2 + (y - cy)^2 = r^2 and ax + by + c = 0 */
    aa = 1 + ((a * a) / (b * b));
    bb = (-2 * cx) - ((2 * a * c) / (b * b)) + ((2 * a * cy) / b);
    cc = (cx * cx) + ((c * c) / (b * b)) + ((2 * c * cy) / b) + (cy * cy) - (r * r);
    delta = (bb * bb) - (4 * aa * cc);

    if (delta > 0)
    {
        x1 = ((-bb) - sqrt(delta)) / (2 * aa);
        x2 = ((-bb) + sqrt(delta)) / (2 * aa);
        y1 = -(((a * x1) + c) / b);
        y2 = -(((a * x2) + c) / b);

        /* consider only solutions inside the road segment */
        if (y1 > 0 && inRoadSegment(x1, y1, cx, cy, r, rs, l))
        {
            td1 = timeOfDeparture(x1, y1, rv, v);
        }
        if (y2 > 0 && inRoadSegment(x2, y2, cx, cy, r, rs, l))
        {
            td2 = timeOfDeparture(x2, y2, rv, v);
        }
    }
    return (td1 > 0) && (td1 < td2) ? td1 : td2;
}

/**
 * @brief Computes the lane departure warning.
 *
 * @param td Time of departure (road-end) (s).
 * @param tdc Time of departure (center-line) (s).
 * @return Lane departure warning.
 */
uint8 alert(float64 td, float64 tdc)
{
    uint8 warning = SEVERITY_NONE;

    if (tdc > 0)
    {
        if (tdc < 0.5)
        {
            warning = SEVERITY_HIGH;
        }
        else if (tdc < 1.0)
        {
            warning = SEVERITY_MEDIUM;
        }
        else if (tdc < 2.0)
        {
            warning = SEVERITY_LOW;
        }
    }
    if (td > 0)
    {
        if (td < 0.5)
        {
            warning = SEVERITY_CRITICAL;
        }
        else if (td < 1.0)
        {
            warning = SEVERITY_HIGH;
        }
        else if (td < 2.0)
        {
            warning = SEVERITY_MEDIUM;
        }
        else if (td < 5.0)
        {
            warning = SEVERITY_LOW;
        }
    }
    return warning;
}

// End of user code
/********************************************************************************************************************
 * End of Section for user's comment
 ********************************************************************************************************************/
FUNC(void, LaneDepartureWarning_CODE) RLaneDepartureWarning (void)
{
    // Start of user code for the runnable of LDW_Controller

    sint16 steeringPosition;
    float64 v;  /* vehicle speed (km/h) */
    float64 cx; /* road curve center (X) */
    float64 cy; /* road curve center (Y) */
    float64 r;  /* road curvature radius */
    float64 l;  /* distance of the vehicle from the front segment–end (m) */
    float64 sx; /* distance of the vehicle from the left road-end (m) */
    float64 dx; /* distance of the vehicle from the right road-end (m) */
    float64 lx; /* distance of the vehicle from the center-line (m) */
    uint8 severity;

    float64 rv; /* vehicle curvature radius */
    float64 vx; /* vehicle curve center (X) */
    float64 vy; /* vehicle curve center (Y) */

    float64 td1; /* time of departure (road-end) */
    float64 tdc; /* time of departure (center-line) */
    float64 td2; /* time of departure (opposite road-end) */

    steeringPosition = Rte_DRead_SteeringPosition_SteeringPosition();
    v = Rte_DRead_VehicleSpeed_VehicleSpeed();
    cx = Rte_DRead_RoadSegment_CurveCenterX();
    cy = Rte_DRead_RoadSegment_CurveCenterY();
    r = Rte_DRead_RoadSegment_CurvatureRadius();
    l = Rte_DRead_RoadSegment_Length();
    sx = Rte_DRead_LanesPosition_RoadEndLeftDistance();
    dx = Rte_DRead_LanesPosition_RoadEndRightDistance();
    lx = Rte_DRead_LanesPosition_OppositeLaneDistance();
    severity = 0;

    rv = WHEELS_DISTANCE/tan(abs(steeringPosition));
    vx = steeringPosition > 0 ? rv : -rv;
    vy = -1;

    /* check availability of lane markings */
    if (v > MINIMUM_OPERATING_SPEED && dx > 0 && sx < 0 && lx > sx && lx < dx)
    {
        if (r == 0)
        {
            if (steeringPosition == 0)
            {
                /* straight steering wheel, straight road */
                td1 = 0;
                tdc = 0;
                td2 = 0;
            }
            else
            {
                /* curving steering wheel, straight road */
                td1 = timeToCurvingCrossLane(vx, vy, rv, dx, l, v);
                tdc = timeToCurvingCrossLane(vx, vy, rv, lx, l, v);
                td2 = timeToCurvingCrossLane(vx, vy, rv, sx, l, v);
            }
        }
        else
        {
            float64 rs;  /* Road curvature radius (vehicle) */
            float64 rc;  /* Road curvature radius (center-line) */
            float64 re;  /* Road curvature radius (road-end) */

            rs = cx > 0 ? r + dx : r + fabs(sx);
            rc = cx > 0 ? r + dx - lx : r + fabs(sx) + lx;
            re = r + dx + fabs(sx);

            if (steeringPosition == 0)
            {
                /* straight steering wheel, curving road */
                td1 = timeToCrossCurvingLane(cx, cy, r, rs, l, v);
                tdc = timeToCrossCurvingLane(cx, cy, rc, rs, l, v);
                td2 = timeToCrossCurvingLane(cx, cy, re, rs, l, v);
            }
            else
            {
                /* curving steering wheel, curving road */
                td1 = timeToCurvingCrossCurvingLane(cx, cy, r, vx, vy, rv, rs, l, v);
                tdc = timeToCurvingCrossCurvingLane(cx, cy, rc, vx, vy, rv, rs, l, v);
                td2 = timeToCurvingCrossCurvingLane(cx, cy, re, vx, vy, rv, rs, l, v);
            }
        }
        severity = alert(td1 > 0 ? td1 : td2, tdc);
    }

    (void) Rte_Write_Warning_Severity(severity);

    // End of user code
}
#define LANEDEPARTUREWARNING_STOP_SEC_CODE
#include "LaneDepartureWarning_MemMap.h"
