
/**************************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2019-2019. All rights reserved.
 * Description: Rte_SpeedControl.c - The file has been not generated and validated in accordance with a safety standard
 * Author: VOS OS
 * Create: 2019-12-31
 *
 **************************************************************************************/

/***************************************************************************
 *
 * Includes
 *
 **************************************************************************/

/*
 * To restrict the SW-C APIs to a runnable use ONE of the following macros
 * #define RTE_RUNNABLEAPI_RPeriodicSpeedControl
 */

/* Include application header file */
#include "Rte_SpeedControl.h"
/* Version verification of the Rte_SpeedControl module */
#if ((RTE_SPEEDCONTROL_AR_RELEASE_MAJOR_VERSION != 4) || (RTE_SPEEDCONTROL_AR_RELEASE_MINOR_VERSION != 4))
#error The imported RTE_SPEEDCONTROL module version is incompatible, there may be problems.
#endif


// Start of user code for user-defined code

// End of user code
/********************************************************************************************************************
 * AtomicSwComponentType : SpeedControl
 * ------------------------------------------------------------------------------------------------------------------
 *
 *******************************************************************************************************************/
#define SPEEDCONTROL_START_SEC_CODE
#include "SpeedControl_MemMap.h"
/********************************************************************************************************************
 * Runnable Entity : RPeriodicSpeedControl
 * ------------------------------------------------------------------------------------------------------------------
 *
 * =======================================
 *           Sender-Receiver
 * =======================================
 * @Function name : Rte_DRead_CC_Active_Active
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(boolean, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_CC_TargetSpeed_TargetSpeed
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(uint8, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_DRead_VehicleSpeed_VehicleSpeed
 * @Description   : Performs an "explicit" read on a sender-receiver communication data element with "data" semantics
 *                  (swImplPolicy != queued). By compatibility, the port may also have a ParameterInterface or a
 *                  NvDataInterface. The Rte_DRead API is used for explicit read by value.
 * @Parameter     : None
 * @Return        : VAR(float64, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_Write_Brake_Brake
 * @Description   : Initiate an "explicit" sender-receiver transmission of data elements with "data" semantic
 *                  (swImplPolicy different from queued).
 * @Parameter     : VAR(float64, AUTOMATIC) value
 * @Return        : VAR(Std_ReturnType, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 * @Function name : Rte_Write_Throttle_Throttle
 * @Description   : Initiate an "explicit" sender-receiver transmission of data elements with "data" semantic
 *                  (swImplPolicy different from queued).
 * @Parameter     : VAR(uint8, AUTOMATIC) value
 * @Return        : VAR(Std_ReturnType, AUTOMATIC)
 * ------------------------------------------------------------------------------------------------------------------
 *
 *******************************************************************************************************************/

/********************************************************************************************************************
 * Start of Section for user's comment
 ********************************************************************************************************************/
// Start of user code : Add user comment for RPeriodicSpeedControl
/* Tuning Parameters of the PID */

#define BRAKE_PARAM 4U
#define THROTTLE_PARAM 2U
#define PID_COEFF 100U

// End of user code
/********************************************************************************************************************
 * End of Section for user's comment
 ********************************************************************************************************************/
FUNC(void, SpeedControl_CODE) RPeriodicSpeedControl (void)
{
    // Start of user code for the runnable of RPeriodicSpeedControl
    
	boolean 	local_CC_Active;
	uint8 		local_CC_TargetSpeed;
	float64		local_VehicleSpeed;
	float64		local_Brake = 0;
	uint8 		local_Throttle = 0;
	
	/* PID Coefficient */
	uint8 B = BRAKE_PARAM;
	uint8 T = THROTTLE_PARAM;
	uint8 Kp = PID_COEFF;
	
	local_CC_Active = Rte_DRead_CC_Active_Active();
	local_VehicleSpeed = Rte_DRead_VehicleSpeed_VehicleSpeed();
	local_CC_TargetSpeed = Rte_DRead_CC_TargetSpeed_TargetSpeed();

	// AG -- fix the broken code...
	Rte_Write_Brake_Brake(0);
	Rte_Write_Throttle_Throttle(0);
        
	if (local_CC_Active) {
		if (local_VehicleSpeed > local_CC_TargetSpeed) {
			local_Brake = B*Kp*(local_VehicleSpeed - local_CC_TargetSpeed);
			(void) Rte_Write_Brake_Brake (local_Brake);
		}
		else if (local_VehicleSpeed < local_CC_TargetSpeed) {
                    float64 tmp = 
			/*local_Throttle =*/ T*Kp*(local_CC_TargetSpeed - local_VehicleSpeed);
                    local_Throttle = tmp >= 255 ? 255 : tmp;
			(void) Rte_Write_Throttle_Throttle (local_Throttle);
		}
        else
        {
            local_Throttle = 0;
            local_Brake = 0;
            /* This case should not happen, add assertion here */
            /* (void) Rte_Write_Throttle_validity (0);*/
        }
    }
    else {
        local_Throttle = 0;
        local_Brake = 0;
        /* This case should not happen, add assertion here */
        /* (void) Rte_Write_Throttle_validity (0);*/
    }
	
    // End of user code
}
#define SPEEDCONTROL_STOP_SEC_CODE
#include "SpeedControl_MemMap.h"
